@@ -20,15 +20,15 @@
 #include <sys/stat.h>
 #include <sys/mman.h>
 #include <pthread.h>
-#ifdef HAVE_ENDIAN_H
-# include <endian.h>
-#endif
 
 #include <asm/types.h>		/* XXX glibc */
 #include "videodev2.h"
 
 #include "grab-ng.h"
 
+#include "struct-dump.h"
+#include "struct-v4l2.h"
+
 /* ---------------------------------------------------------------------- */
 
 /* open+close */
@@ -49,7 +49,7 @@ static int   v4l2_overlay(void *handle, struct ng_video_fmt *fmt, int x, int y,
 
 /* capture video */
 static int v4l2_setformat(void *handle, struct ng_video_fmt *fmt);
-static int v4l2_startvideo(void *handle, int fps, int buffers);
+static int v4l2_startvideo(void *handle, int fps, unsigned int buffers);
 static void v4l2_stopvideo(void *handle);
 static struct ng_video_buf* v4l2_nextframe(void *handle);
 static struct ng_video_buf* v4l2_getimage(void *handle);
@@ -61,7 +61,7 @@ static int v4l2_tuned(void *handle);
 
 /* ---------------------------------------------------------------------- */
 
-#define WANTED_BUFFERS 16
+#define WANTED_BUFFERS 32
 
 #define MAX_INPUT   16
 #define MAX_NORM    16
@@ -76,7 +76,7 @@ struct v4l2_handle {
     struct v4l2_capability	cap;
     struct v4l2_streamparm	streamparm;
     struct v4l2_input		inp[MAX_INPUT];
-    struct v4l2_enumstd		std[MAX_NORM];
+    struct v4l2_standard      	std[MAX_NORM];
     struct v4l2_fmtdesc		fmt[MAX_FORMAT];
     struct v4l2_queryctrl	ctl[MAX_CTRL*2];
 
@@ -92,12 +92,15 @@ struct v4l2_handle {
     struct v4l2_requestbuffers     reqbufs;
     struct v4l2_buffer             buf_v4l2[WANTED_BUFFERS];
     struct ng_video_buf            buf_me[WANTED_BUFFERS];
-    int                            queue,waiton;
+    unsigned int                   queue,waiton;
 
     /* overlay */
     struct v4l2_framebuffer        ov_fb;
-    struct v4l2_window             ov_win;
+    struct v4l2_format             ov_win;
     struct v4l2_clip               ov_clips[256];
+#if 0
+    enum v4l2_field                ov_fields;
+#endif
     int                            ov_error;
     int                            ov_enabled;
     int                            ov_on;
@@ -129,22 +132,19 @@ struct ng_vid_driver v4l2_driver = {
 };
 
 static __u32 xawtv_pixelformat[VIDEO_FMT_COUNT] = {
-    0,                    /* unused   */
-    V4L2_PIX_FMT_HI240,   /* RGB8     */
-    V4L2_PIX_FMT_GREY,    /* GRAY8    */
-    V4L2_PIX_FMT_RGB555,  /* RGB15_LE */
-    V4L2_PIX_FMT_RGB565,  /* RGB16_LE */
-    V4L2_PIX_FMT_RGB555X, /* RGB15_BE */
-    V4L2_PIX_FMT_RGB565X, /* RGB16_BE */
-    V4L2_PIX_FMT_BGR24,   /* BGR24    */
-    V4L2_PIX_FMT_BGR32,   /* BGR32    */
-    V4L2_PIX_FMT_RGB24,   /* RGB24    */
-    0,                    /* RGB32    */
-    0,                    /* LUT 2    */
-    0,                    /* LUT 4    */
-    V4L2_PIX_FMT_YUYV,    /* YUV422   */
-    V4L2_PIX_FMT_YUV422P, /* YUV422P  */
-    V4L2_PIX_FMT_YUV420,  /* YUV420P  */
+    [ VIDEO_RGB08 ]    = V4L2_PIX_FMT_HI240,
+    [ VIDEO_GRAY ]     = V4L2_PIX_FMT_GREY,
+    [ VIDEO_RGB15_LE ] = V4L2_PIX_FMT_RGB555,
+    [ VIDEO_RGB16_LE ] = V4L2_PIX_FMT_RGB565,
+    [ VIDEO_RGB15_BE ] = V4L2_PIX_FMT_RGB555X,
+    [ VIDEO_RGB16_BE ] = V4L2_PIX_FMT_RGB565X,
+    [ VIDEO_BGR24 ]    = V4L2_PIX_FMT_BGR24,
+    [ VIDEO_BGR32 ]    = V4L2_PIX_FMT_BGR32,
+    [ VIDEO_RGB24 ]    = V4L2_PIX_FMT_RGB24,
+    [ VIDEO_YUYV ]     = V4L2_PIX_FMT_YUYV,
+    [ VIDEO_UYVY ]     = V4L2_PIX_FMT_UYVY,
+    [ VIDEO_YUV422P ]  = V4L2_PIX_FMT_YUV422P,
+    [ VIDEO_YUV420P ]  = V4L2_PIX_FMT_YUV420,
 };
 
 static struct STRTAB stereo[] = {
@@ -160,22 +160,6 @@ static struct STRTAB stereo[] = {
 
 #define PREFIX "ioctl: "
 
-static const char *io_names[] = {
-    "QUERYCAP", "1", "ENUM_PIXFMT", "ENUM_FBUFFMT", "G_FMT", "S_FMT",
-    "G_COMP", "S_COMP", "REQBUFS", "QUERYBUF", "G_FBUF", "S_FBUF",
-    "G_WIN", "S_WIN", "PREVIEW", "QBUF", "16", "DQBUF", "STREAMON",
-    "STREAMOFF", "G_PERF", "G_PARM", "S_PARM", "G_STD", "S_STD",
-    "ENUMSTD", "ENUMINPUT", "G_CTRL", "S_CTRL", "G_TUNER", "S_TUNER",
-    "G_FREQ", "S_FREQ", "G_AUDIO", "S_AUDIO", "35", "QUERYCTRL",
-    "QUERYMENU", "G_INPUT", "S_INPUT", "ENUMCVT", "41", "42", "43",
-    "44", "45",  "G_OUTPUT", "S_OUTPUT", "ENUMOUTPUT", "G_AUDOUT",
-    "S_AUDOUT", "ENUMFX", "G_EFFECT", "S_EFFECT", "G_MODULATOR",
-    "S_MODULATOR"
-};
-static const int io_count = (sizeof(io_names)/sizeof(char*));
-#define IONAME(cmd)	((cmd & 0xff) < io_count ? \
-			io_names[cmd & 0xff] : "UNKNOWN")
-
 static int
 xioctl(int fd, int cmd, void *arg, int mayfail)
 {
@@ -186,277 +170,27 @@ xioctl(int fd, int cmd, void *arg, int mayfail)
 	return rc;
     if (mayfail && errno == mayfail && ng_debug < 2)
 	return rc;
-    switch (cmd) {
-    case VIDIOC_QUERYCAP:
-    {
-	struct v4l2_capability *a = arg;
-	fprintf(stderr,PREFIX "VIDIOC_QUERYCAP(%s,type=0x%x,in=%d,out=%d,"
-		"audio=%d,size=%dx%d-%dx%d,fps=%d,flags=0x%x)",
-		a->name,a->type,a->inputs,a->outputs,a->audios,
-		a->minwidth,a->minheight,a->maxwidth,a->maxheight,
-		a->maxframerate,a->flags);
-	break;
-    }
-    
-    case VIDIOC_G_FMT:
-    case VIDIOC_S_FMT:
-    {
-	struct v4l2_format *a = arg;
-
-	fprintf(stderr,PREFIX "VIDIOC_%s(type=%d,",IONAME(cmd),a->type);
-	switch (a->type) {
-	case V4L2_BUF_TYPE_CAPTURE:
-	    fprintf(stderr,
-		    "%dx%d,depth=%d,%c%c%c%c,flags=0x%x,bpl=%d,size=%d)",
-		    a->fmt.pix.width,a->fmt.pix.height,a->fmt.pix.depth,
-		    a->fmt.pix.pixelformat & 0xff,
-		    (a->fmt.pix.pixelformat >>  8) & 0xff,
-		    (a->fmt.pix.pixelformat >> 16) & 0xff,
-		    (a->fmt.pix.pixelformat >> 24) & 0xff,
-		    a->fmt.pix.depth,a->fmt.pix.bytesperline,
-		    a->fmt.pix.sizeimage);
-	    break;
-	default:
-	    fprintf(stderr,"??" "?)"); /* break trigraph */
-	    break;
-	}
-	break;
-    }
-    case VIDIOC_REQBUFS:
-    {
-	struct v4l2_requestbuffers *a = arg;
-	
-	fprintf(stderr,PREFIX "VIDIOC_REQBUFS(count=%d,type=%d)",
-		a->count,a->type);
-	break;
-    }
-    case VIDIOC_QBUF:
-    case VIDIOC_DQBUF:
-    {
-	struct v4l2_buffer *a = arg;
-	
-	fprintf(stderr,PREFIX "VIDIOC_%s(%d,type=%d,off=%d,len=%d,used=%d,"
-		"flags=0x%x,ts=%Ld,seq=%d)",
-		IONAME(cmd),a->index,a->type,a->offset,a->length,
-		a->bytesused,a->flags,a->timestamp,a->sequence);
-	break;
-    }
-    
-    case VIDIOC_G_WIN:
-    case VIDIOC_S_WIN:
-    {
-	struct v4l2_window *a = arg;
-	
-	fprintf(stderr,PREFIX "VIDIOC_%s(%dx%d+%d+%d,key=0x%x,clips=%d)",
-		IONAME(cmd), a->width, a->height, a->x, a->y,
-		a->chromakey,a->clipcount);
-	break;
-    }
-    case VIDIOC_PREVIEW:
-    {
-	int *a = arg;
-
-	fprintf(stderr,PREFIX "VIDIOC_PREVIEW(%s)",*a ? "on" : "off");
-	break;
-    }
-    
-    case VIDIOC_QUERYCTRL:
-    {
-	struct v4l2_queryctrl *a = arg;
-	
-	fprintf(stderr,PREFIX "VIDIOC_QUERYCTRL(id=%d,%s,%d-%d/%d,def=%d,"
-		"type=%d,flags=0x%x)",
-		a->id,a->name,a->minimum,a->maximum,a->step,
-		a->default_value,a->type,a->flags);
-	break;
-    }
-    case VIDIOC_QUERYMENU:
-    {
-	struct v4l2_querymenu *a = arg;
-
-	fprintf(stderr,PREFIX "VIDIOC_QUERYMENU(id=%d,index=%d,%s)",
-		a->id,a->index,a->name);
-	break;
-    }
-    case VIDIOC_G_CTRL:
-    case VIDIOC_S_CTRL:
-    {
-	struct v4l2_control *a = arg;
-	
-	fprintf(stderr,PREFIX "VIDIOC_%s(id=%d,value=%d)",
-		IONAME(cmd),a->id,a->value);
-	break;
-    }
-    
-    default:
-	fprintf(stderr,PREFIX "VIDIOC_%s(cmd=0x%x)",IONAME(cmd),cmd);
-	break;
-    }
+    print_ioctl(stderr,ioctls_v4l2,PREFIX,cmd,arg);
     fprintf(stderr,": %s\n",(rc == 0) ? "ok" : strerror(errno));
     return rc;
 }
 
-static void
-print_bits(char *title, char **names, int count, int value)
-{
-    int i;
-    
-    fprintf(stderr,"%s: ",title);
-    for (i = 0; i < count; i++) {
-	if (value & (1 << i))
-	    fprintf(stderr,"%s ",names[i]);
-    }
-    fprintf(stderr,"\n");
-}    
-
-static void
-print_device_capabilities(struct v4l2_handle *h)
-{
-    static char *cap_type[] = {
-	"capture",
-	"codec",
-	"output",
-	"fx",
-	"vbi",
-	"vtr",
-	"vtx",
-	"radio",
-    };
-    static char *cap_flags[] = {
-	"read",
-	"write",
-	"streaming",
-	"preview",
-	"select",
-	"tuner",
-	"monochrome",
-	"teletext"
-    };
-    static char *ctl_type[] = {
-	"integer",
-	"boolean",
-	"menu"
-    };
-    static char *cap_parm[] = {
-	"highquality",
-	"vflip",
-	"hflip"
-    };
-
-    int i;
-
-    fprintf(stderr,"\n*** v4l2: video device capabilities ***\n");
-
-    /* capabilities */
-    fprintf(stderr,"type: %s\n",
-	    h->cap.type < sizeof(cap_type)/sizeof(char*) ?
-	    cap_type[h->cap.type] : "unknown");
-    print_bits("flags",cap_flags,sizeof(cap_flags)/sizeof(char*),h->cap.flags);
-    fprintf(stderr,"\n");
-    fprintf(stderr,"inputs: %d\naudios: %d\n",h->cap.inputs,h->cap.audios);
-    fprintf(stderr,"size: %dx%d => %dx%d\n",
-	    h->cap.minwidth,h->cap.minheight,h->cap.maxwidth,h->cap.maxheight);
-    fprintf(stderr,"fps: %d max\n",h->cap.maxframerate);
-
-    /* inputs */
-    fprintf(stderr,"video inputs:\n");
-    for (i = 0; i < h->ninputs; i++) {
-	printf("  %d: \"%s\", tuner: %s, audio: %s\n", i, h->inp[i].name,
-	       (h->inp[i].type       == V4L2_INPUT_TYPE_TUNER) ? "yes" : "no",
-	       (h->inp[i].capability &  V4L2_INPUT_CAP_AUDIO)  ? "yes" : "no");
-    }
-
-    /* video standards */
-    fprintf(stderr,"video standards:\n");
-    for (i = 0; i < h->nstds; i++) {
-	printf("  %d: \"%s\"\n", i, h->std[i].std.name);
-    }
-
-    /* capture formats */
-    fprintf(stderr,"capture formats:\n");
-    for (i = 0; i < h->nfmts; i++) {
-	fprintf(stderr,"  %d: %c%c%c%c, depth=%d,%s \"%s\"\n", i,
-		h->fmt[i].pixelformat & 0xff,
-		(h->fmt[i].pixelformat >>  8) & 0xff,
-		(h->fmt[i].pixelformat >> 16) & 0xff,
-		(h->fmt[i].pixelformat >> 24) & 0xff,
-		h->fmt[i].depth,
-		(h->fmt[i].flags & V4L2_FMT_FLAG_COMPRESSED) ? " compressed" : "",
-		h->fmt[i].description);
-    }
-
-    /* capture parameters */
-    fprintf(stderr,"capture parameters:\n");
-    print_bits("  cap",cap_parm,sizeof(cap_parm)/sizeof(char*),
-	       h->streamparm.parm.capture.capability);
-    print_bits("  cur",cap_parm,sizeof(cap_parm)/sizeof(char*),
-	       h->streamparm.parm.capture.capturemode);
-    fprintf(stderr,"  timeperframe=%ld\n",
-	    h->streamparm.parm.capture.timeperframe);
-
-    /* controls */
-    fprintf(stderr,"supported controls:\n");
-    for (i = 0; i < MAX_CTRL*2; i++) {
-	if (h->ctl[i].id == -1)
-	    continue;
-	fprintf(stderr,"  %2d: \"%s\", [%d .. %d], step=%d, def=%d, type=%s\n",
-		i, h->ctl[i].name,
-		h->ctl[i].minimum,h->ctl[i].maximum,
-		h->ctl[i].step,h->ctl[i].default_value,
-		ctl_type[h->ctl[i].type]);
-    }
-    fprintf(stderr,"\n");
-}
-
 static void
 print_bufinfo(struct v4l2_buffer *buf)
 {
     static char *type[] = {
-	"",
-	"capture",
-	"codec in",
-	"codec out",
-	"effects in1",
-	"effects in2",
-	"effects out",
-	"video out"
+	[V4L2_BUF_TYPE_VIDEO_CAPTURE] = "video-cap",
+	[V4L2_BUF_TYPE_VIDEO_OVERLAY] = "video-over",
+	[V4L2_BUF_TYPE_VIDEO_OUTPUT]  = "video-out",
+	[V4L2_BUF_TYPE_VBI_CAPTURE]   = "vbi-cap",
+	[V4L2_BUF_TYPE_VBI_OUTPUT]    = "vbi-out",
     };
 
     fprintf(stderr,"v4l2: buf %d: %s 0x%x+%d, used %d\n",
-		   buf->index,
-	    	   buf->type < sizeof(type)/sizeof(char*) ?
-			type[buf->type] : "unknown",
-		   buf->offset,buf->length,buf->bytesused);
-}
-
-static void
-print_fbinfo(struct v4l2_framebuffer *fb)
-{
-    static char *fb_cap[] = {
-	"extern",
-	"chromakey",
-	"clipping",
-	"scale-up",
-	"scale-down"
-    };
-    static char *fb_flags[] = {
-	"primary",
-	"overlay",
-	"chromakey"
-    };
-
-    /* capabilities */
-    fprintf(stderr,"v4l2: framebuffer info\n");
-    print_bits("  cap",fb_cap,sizeof(fb_cap)/sizeof(char*),fb->capability);
-    print_bits("  flags",fb_cap,sizeof(fb_flags)/sizeof(char*),fb->flags);
-    fprintf(stderr,"  base: %p %p %p\n",fb->base[0],fb->base[1],fb->base[2]);
-    fprintf(stderr,"  format: %dx%d, %c%c%c%c, %d byte\n",
-	    fb->fmt.width, fb->fmt.height,
-	    fb->fmt.pixelformat & 0xff,
-	    (fb->fmt.pixelformat >>  8) & 0xff,
-	    (fb->fmt.pixelformat >> 16) & 0xff,
-	    (fb->fmt.pixelformat >> 24) & 0xff,
-	    fb->fmt.sizeimage);
+	    buf->index,
+	    buf->type < sizeof(type)/sizeof(char*)
+	    ? type[buf->type] : "unknown",
+	    buf->m.offset,buf->length,buf->bytesused);
 }
 
 /* ---------------------------------------------------------------------- */
@@ -467,9 +201,9 @@ get_device_capabilities(struct v4l2_handle *h)
 {
     int i;
     
-    for (h->ninputs = 0; h->ninputs < h->cap.inputs; h->ninputs++) {
+    for (h->ninputs = 0; h->ninputs < MAX_INPUT; h->ninputs++) {
 	h->inp[h->ninputs].index = h->ninputs;
-	if (-1 == xioctl(h->fd, VIDIOC_ENUMINPUT, &h->inp[h->ninputs], 0))
+	if (-1 == xioctl(h->fd, VIDIOC_ENUMINPUT, &h->inp[h->ninputs], EINVAL))
 	    break;
     }
     for (h->nstds = 0; h->nstds < MAX_NORM; h->nstds++) {
@@ -479,11 +213,12 @@ get_device_capabilities(struct v4l2_handle *h)
     }
     for (h->nfmts = 0; h->nfmts < MAX_FORMAT; h->nfmts++) {
 	h->fmt[h->nfmts].index = h->nfmts;
-	if (-1 == xioctl(h->fd, VIDIOC_ENUM_PIXFMT, &h->fmt[h->nfmts], EINVAL))
+	h->fmt[h->nfmts].type  = V4L2_BUF_TYPE_VIDEO_CAPTURE;
+	if (-1 == xioctl(h->fd, VIDIOC_ENUM_FMT, &h->fmt[h->nfmts], EINVAL))
 	    break;
     }
 
-    h->streamparm.type = V4L2_BUF_TYPE_CAPTURE;
+    h->streamparm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
     ioctl(h->fd,VIDIOC_G_PARM,&h->streamparm);
 
     /* controls */
@@ -510,7 +245,7 @@ build_norms(struct v4l2_handle *h)
     norms = malloc(sizeof(struct STRTAB) * (h->nstds+1));
     for (i = 0; i < h->nstds; i++) {
 	norms[i].nr  = i;
-	norms[i].str = h->std[i].std.name;
+	norms[i].str = h->std[i].name;
     }
     norms[i].nr  = -1;
     norms[i].str = NULL;
@@ -536,8 +271,8 @@ build_inputs(struct v4l2_handle *h)
 /* ---------------------------------------------------------------------- */
 
 static struct V4L2_ATTR {
-    int id;
-    int v4l2;
+    unsigned int id;
+    unsigned int v4l2;
 } v4l2_attr[] = {
     { ATTR_ID_VOLUME,   V4L2_CID_AUDIO_VOLUME },
     { ATTR_ID_MUTE,     V4L2_CID_AUDIO_MUTE   },
@@ -576,7 +311,7 @@ v4l2_add_attr(struct v4l2_handle *h, struct v4l2_queryctrl *ctl,
 	      int id, struct STRTAB *choices)
 {
     static int private_ids = ATTR_ID_COUNT;
-    int i;
+    unsigned int i;
     
     h->attr = realloc(h->attr,(h->nattr+2) * sizeof(struct ng_attribute));
     memset(h->attr+h->nattr,0,sizeof(struct ng_attribute)*2);
@@ -612,6 +347,8 @@ v4l2_add_attr(struct v4l2_handle *h, struct v4l2_queryctrl *ctl,
     } else {
 	/* for norms + inputs */
 	h->attr[h->nattr].id      = id;
+	if (-1 == h->attr[h->nattr].id)
+	    h->attr[h->nattr].id  = private_ids++;
 	h->attr[h->nattr].defval  = 0;
 	h->attr[h->nattr].type    = ATTR_TYPE_CHOICE;
 	h->attr[h->nattr].choices = choices;
@@ -631,7 +368,9 @@ static int v4l2_read_attr(struct ng_attribute *attr)
     const struct v4l2_queryctrl *ctl = attr->priv;
     struct v4l2_control c;
     struct v4l2_tuner tuner;
+    v4l2_std_id std;
     int value = 0;
+    int i;
 
     if (NULL != ctl) {
 	c.id = ctl->id;
@@ -639,7 +378,11 @@ static int v4l2_read_attr(struct ng_attribute *attr)
 	value = c.value;
 	
     } else if (attr->id == ATTR_ID_NORM) {
-	value = -1; /* FIXME */
+	value = -1;
+	xioctl(h->fd,VIDIOC_G_STD,&std,0);
+	for (i = 0; i < h->nstds; i++)
+	    if (std & h->std[i].id)
+		value = i;
 	
     } else if (attr->id == ATTR_ID_INPUT) {
 	xioctl(h->fd,VIDIOC_G_INPUT,&value,0);
@@ -683,7 +426,7 @@ static void v4l2_write_attr(struct ng_attribute *attr, int value)
 	xioctl(h->fd,VIDIOC_S_CTRL,&c,0);
 	
     } else if (attr->id == ATTR_ID_NORM) {
-	xioctl(h->fd,VIDIOC_S_STD,&h->std[value].std,0);
+	xioctl(h->fd,VIDIOC_S_STD,&h->std[value].id,0);
 	
     } else if (attr->id == ATTR_ID_INPUT) {
 	xioctl(h->fd,VIDIOC_S_INPUT,&value,0);
@@ -709,30 +452,33 @@ v4l2_open(char *device)
 	return NULL;
     memset(h,0,sizeof(*h));
     
-    if (-1 == (h->fd = open(device,O_RDWR))) {
+    if (-1 == (h->fd = open(device, O_RDWR))) {
 	fprintf(stderr,"v4l2: open %s: %s\n",device,strerror(errno));
 	goto err;
     }
 
-    if (-1 == ioctl(h->fd,VIDIOC_QUERYCAP,&h->cap))
+    if (-1 == xioctl(h->fd,VIDIOC_QUERYCAP,&h->cap,EINVAL))
 	goto err;
     if (ng_debug)
 	fprintf(stderr, "v4l2: open\n");
     fcntl(h->fd,F_SETFD,FD_CLOEXEC);
     if (ng_debug)
-	fprintf(stderr,"v4l2: device is %s\n",h->cap.name);
-
+	fprintf(stderr,"v4l2: device info:\n"
+		"  %s %d.%d.%d / %s @ %s\n",
+		h->cap.driver,
+		(h->cap.version >> 16) & 0xff,
+		(h->cap.version >>  8) & 0xff,
+		h->cap.version         & 0xff,
+		h->cap.card,h->cap.bus_info);
     get_device_capabilities(h);
-    if (ng_debug)
-	print_device_capabilities(h);
 
     /* attributes */
     v4l2_add_attr(h, NULL, ATTR_ID_NORM,  build_norms(h));
     v4l2_add_attr(h, NULL, ATTR_ID_INPUT, build_inputs(h));
-    if (h->cap.flags & V4L2_FLAG_TUNER)
+    if (h->cap.capabilities & V4L2_CAP_TUNER)
 	v4l2_add_attr(h, NULL, ATTR_ID_AUDIO_MODE, stereo);
     for (i = 0; i < MAX_CTRL*2; i++) {
-	if (h->ctl[i].id == -1)
+	if (h->ctl[i].id == UNSET)
 	    continue;
 	v4l2_add_attr(h, &h->ctl[i], 0, NULL);
     }
@@ -770,7 +516,7 @@ static char*
 v4l2_devname(void *handle)
 {
     struct v4l2_handle *h = handle;
-    return h->cap.name;
+    return h->cap.card;
 }
 
 static int v4l2_flags(void *handle)
@@ -778,12 +524,11 @@ static int v4l2_flags(void *handle)
     struct v4l2_handle *h = handle;
     int ret = 0;
 
-    if (h->cap.flags & V4L2_FLAG_PREVIEW && !h->ov_error)
+    if (h->cap.capabilities & V4L2_CAP_VIDEO_OVERLAY && !h->ov_error)
 	ret |= CAN_OVERLAY;
-    if ((h->cap.flags & V4L2_FLAG_STREAMING) ||
-	(h->cap.flags & V4L2_FLAG_READ))
+    if (h->cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)
 	ret |= CAN_CAPTURE;
-    if (h->cap.flags & V4L2_FLAG_TUNER)
+    if (h->cap.capabilities & V4L2_CAP_TUNER)
 	ret |= CAN_TUNE;
     return ret;
 }
@@ -800,20 +545,25 @@ static unsigned long
 v4l2_getfreq(void *handle)
 {
     struct v4l2_handle *h = handle;
-    unsigned long freq;
+    struct v4l2_frequency f;
 
-    xioctl(h->fd, VIDIOC_G_FREQ, &freq, 0);
-    return freq;
+    memset(&f,0,sizeof(f));
+    xioctl(h->fd, VIDIOC_G_FREQUENCY, &f, 0);
+    return f.frequency;
 }
 
 static void
 v4l2_setfreq(void *handle, unsigned long freq)
 {
     struct v4l2_handle *h = handle;
+    struct v4l2_frequency f;
 
     if (ng_debug)
 	fprintf(stderr,"v4l2: freq: %.3f\n",(float)freq/16);
-    xioctl(h->fd, VIDIOC_S_FREQ, &freq, 0);
+    memset(&f,0,sizeof(f));
+    f.type = V4L2_TUNER_ANALOG_TV;
+    f.frequency = freq;
+    xioctl(h->fd, VIDIOC_S_FREQUENCY, &f, 0);
 }
 
 static int
@@ -823,6 +573,7 @@ v4l2_tuned(void *handle)
     struct v4l2_tuner tuner;
 
     usleep(10000);
+    memset(&tuner,0,sizeof(tuner));
     if (-1 == xioctl(h->fd,VIDIOC_G_TUNER,&tuner,0))
 	return 0;
     return tuner.signal ? 1 : 0;
@@ -839,11 +590,8 @@ v4l2_setupfb(void *handle, struct ng_video_fmt *fmt, void *base)
     if (-1 == xioctl(h->fd, VIDIOC_G_FBUF, &h->ov_fb, 0))
 	return -1;
     
-    if (1 /* ng_debug */)
-	print_fbinfo(&h->ov_fb);
-
     /* double-check settings */
-    if (NULL != base && h->ov_fb.base[0] != base) {
+    if (NULL != base && h->ov_fb.base != base) {
 	fprintf(stderr,"v4l2: WARNING: framebuffer base address mismatch\n");
 	fprintf(stderr,"v4l2: me=%p v4l=%p\n",base,h->ov_fb.base);
 	h->ov_error = 1;
@@ -857,8 +605,7 @@ v4l2_setupfb(void *handle, struct ng_video_fmt *fmt, void *base)
 	h->ov_error = 1;
 	return -1;
     }
-    if ((h->ov_fb.fmt.flags & V4L2_FMT_FLAG_BYTESPERLINE) &&
-	fmt->bytesperline >  0 &&
+    if (fmt->bytesperline > 0 &&
 	fmt->bytesperline != h->ov_fb.fmt.bytesperline) {
 	fprintf(stderr,"v4l2: WARNING: framebuffer bpl mismatch\n");
 	fprintf(stderr,"v4l2: me=%d v4l=%d\n",
@@ -891,6 +638,7 @@ v4l2_overlay(void *handle, struct ng_video_fmt *fmt, int x, int y,
 	     struct OVERLAY_CLIP *oc, int count, int aspect)
 {
     struct v4l2_handle *h = handle;
+    struct v4l2_format win;
     int rc,i;
 
     if (h->ov_error)
@@ -902,7 +650,7 @@ v4l2_overlay(void *handle, struct ng_video_fmt *fmt, int x, int y,
 	if (h->ov_enabled) {
 	    h->ov_enabled = 0;
 	    h->ov_on = 0;
-	    xioctl(h->fd, VIDIOC_PREVIEW, &h->ov_on, 0);
+	    xioctl(h->fd, VIDIOC_OVERLAY, &h->ov_on, 0);
 	}
 	return 0;
     }
@@ -910,40 +658,39 @@ v4l2_overlay(void *handle, struct ng_video_fmt *fmt, int x, int y,
     if (ng_debug)
 	fprintf(stderr,"v4l2: overlay win=%dx%d+%d+%d, %d clips\n",
 		fmt->width,fmt->height,x,y,count);
-    h->ov_win.x          = x;
-    h->ov_win.y          = y;
-    h->ov_win.width      = fmt->width;
-    h->ov_win.height     = fmt->height;
+    memset(&win,0,sizeof(win));
+    win.type = V4L2_BUF_TYPE_VIDEO_OVERLAY;
+    win.fmt.win.w.left    = x;
+    win.fmt.win.w.top     = y;
+    win.fmt.win.w.width   = fmt->width;
+    win.fmt.win.w.height  = fmt->height;
 
     /* check against max. size */
-    ioctl(h->fd,VIDIOC_QUERYCAP,&h->cap);
-    if (h->ov_win.width > h->cap.maxwidth) {
-	h->ov_win.width = h->cap.maxwidth;
-	h->ov_win.x += (fmt->width - h->ov_win.width)/2;
-    }
-    if (h->ov_win.height > h->cap.maxheight) {
-	h->ov_win.height = h->cap.maxheight;
-	h->ov_win.y +=  (fmt->height - h->ov_win.height)/2;
-    }
+    xioctl(h->fd,VIDIOC_TRY_FMT,&win,0);
+    if (win.fmt.win.w.width != (int)fmt->width)
+	win.fmt.win.w.left = x + (fmt->width - win.fmt.win.w.width)/2;
+    if (win.fmt.win.w.height != (int)fmt->height)
+	win.fmt.win.w.top = y + (fmt->height - win.fmt.win.w.height)/2;
     if (aspect)
-	ng_ratio_fixup(&h->ov_win.width,&h->ov_win.height,
-		       &h->ov_win.x,&h->ov_win.y);
+	ng_ratio_fixup(&win.fmt.win.w.width,&win.fmt.win.w.height,
+		       &win.fmt.win.w.left,&win.fmt.win.w.top);
 
     /* fixups */
-    ng_check_clipping(h->ov_win.width, h->ov_win.height,
-		      x - h->ov_win.x, y - h->ov_win.y,
+    ng_check_clipping(win.fmt.win.w.width, win.fmt.win.w.height,
+		      x - win.fmt.win.w.left, y - win.fmt.win.w.top,
 		      oc, &count);
 
-    if (h->ov_fb.capability & V4L2_FBUF_CAP_CLIPPING) {
-	h->ov_win.clips      = h->ov_clips;
-	h->ov_win.clipcount  = count;
+    h->ov_win = win;
+    if (h->ov_fb.capability & V4L2_FBUF_CAP_LIST_CLIPPING) {
+	h->ov_win.fmt.win.clips      = h->ov_clips;
+	h->ov_win.fmt.win.clipcount  = count;
 	
 	for (i = 0; i < count; i++) {
-	    h->ov_clips[i].next   = (i+1 == count) ? NULL : &h->ov_clips[i+1];
-	    h->ov_clips[i].x      = oc[i].x1;
-	    h->ov_clips[i].y      = oc[i].y1;
-	    h->ov_clips[i].width  = oc[i].x2-oc[i].x1;
-	    h->ov_clips[i].height = oc[i].y2-oc[i].y1;
+	    h->ov_clips[i].next = (i+1 == count) ? NULL : &h->ov_clips[i+1];
+	    h->ov_clips[i].c.left   = oc[i].x1;
+	    h->ov_clips[i].c.top    = oc[i].y1;
+	    h->ov_clips[i].c.width  = oc[i].x2-oc[i].x1;
+	    h->ov_clips[i].c.height = oc[i].y2-oc[i].y1;
 	}
     }
 #if 0
@@ -951,11 +698,11 @@ v4l2_overlay(void *handle, struct ng_video_fmt *fmt, int x, int y,
 	h->ov_win.chromakey  = 0;    /* FIXME */
     }
 #endif
-    rc = xioctl(h->fd, VIDIOC_S_WIN, &h->ov_win, 0);
+    rc = xioctl(h->fd, VIDIOC_S_FMT, &h->ov_win, 0);
 
     h->ov_enabled = (0 == rc) ? 1 : 0;
     h->ov_on      = (0 == rc) ? 1 : 0;
-    xioctl(h->fd, VIDIOC_PREVIEW, &h->ov_on, 0);
+    xioctl(h->fd, VIDIOC_OVERLAY, &h->ov_on, 0);
 
     return 0;
 }
@@ -1002,7 +749,7 @@ v4l2_waiton(struct v4l2_handle *h)
     
     /* wait for the next frame */
  again:
-    tv.tv_sec  = 1;
+    tv.tv_sec  = 5;
     tv.tv_usec = 0;
     FD_ZERO(&rdset);
     FD_SET(h->fd, &rdset);
@@ -1019,11 +766,30 @@ v4l2_waiton(struct v4l2_handle *h)
 
     /* get it */
     memset(&buf,0,sizeof(buf));
-    buf.type = V4L2_BUF_TYPE_CAPTURE;
+    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
     if (-1 == xioctl(h->fd,VIDIOC_DQBUF,&buf, 0))
 	return -1;
     h->waiton++;
     h->buf_v4l2[buf.index] = buf;
+
+#if 0
+    if (1) {
+	/* for driver debugging */
+	static const char *fn[] = {
+		"any", "none", "top", "bottom",
+		"interlaced", "tb", "bt", "alternate",
+	};
+	static struct timeval last;
+	signed long  diff;
+
+	diff  = (buf.timestamp.tv_sec - last.tv_sec) * 1000000;
+	diff += buf.timestamp.tv_usec - last.tv_usec;
+	fprintf(stderr,"\tdiff %6.1f ms  buf %d  field %d [%s]\n",
+		diff/1000.0, buf.index, buf.field, fn[buf.field%8]);
+	last = buf.timestamp;
+    }
+#endif
+
     return buf.index;
 }
 
@@ -1031,25 +797,27 @@ static int
 v4l2_start_streaming(struct v4l2_handle *h, int buffers)
 {
     int disable_overlay = 0;
-    int i;
+    unsigned int i;
     
     /* setup buffers */
-    h->reqbufs.count = buffers;
-    h->reqbufs.type  = V4L2_BUF_TYPE_CAPTURE;
+    h->reqbufs.count  = buffers;
+    h->reqbufs.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
+    h->reqbufs.memory = V4L2_MEMORY_MMAP;
     if (-1 == xioctl(h->fd, VIDIOC_REQBUFS, &h->reqbufs, 0))
 	return -1;
     for (i = 0; i < h->reqbufs.count; i++) {
-	h->buf_v4l2[i].index = i;
-	h->buf_v4l2[i].type  = V4L2_BUF_TYPE_CAPTURE;
-	if (-1 == ioctl(h->fd, VIDIOC_QUERYBUF, &h->buf_v4l2[i]))
+	h->buf_v4l2[i].index  = i;
+	h->buf_v4l2[i].type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
+	h->buf_v4l2[i].memory = V4L2_MEMORY_MMAP;
+	if (-1 == xioctl(h->fd, VIDIOC_QUERYBUF, &h->buf_v4l2[i], 0))
 	    return -1;
 	h->buf_me[i].fmt  = h->fmt_me;
 	h->buf_me[i].size = h->buf_me[i].fmt.bytesperline *
 	    h->buf_me[i].fmt.height;
 	h->buf_me[i].data = mmap(NULL, h->buf_v4l2[i].length,
 				 PROT_READ | PROT_WRITE, MAP_SHARED,
-				 h->fd, h->buf_v4l2[i].offset);
-	if ((void*)-1 == h->buf_me[i].data) {
+				 h->fd, h->buf_v4l2[i].m.offset);
+	if (MAP_FAILED == h->buf_me[i].data) {
 	    perror("mmap");
 	    return -1;
 	}
@@ -1064,7 +832,7 @@ v4l2_start_streaming(struct v4l2_handle *h, int buffers)
     /* turn off preview (if needed) */
     if (disable_overlay) {
 	h->ov_on = 0;
-	xioctl(h->fd, VIDIOC_PREVIEW, &h->ov_on, 0);
+	xioctl(h->fd, VIDIOC_OVERLAY, &h->ov_on, 0);
 	if (ng_debug)
 	    fprintf(stderr,"v4l2: overlay off (start_streaming)\n");
     }
@@ -1084,7 +852,7 @@ v4l2_start_streaming(struct v4l2_handle *h, int buffers)
 static void
 v4l2_stop_streaming(struct v4l2_handle *h)
 {
-    int i;
+    unsigned int i;
     
     /* stop capture */
     if (-1 == ioctl(h->fd,VIDIOC_STREAMOFF,&h->fmt_v4l2.type))
@@ -1094,6 +862,8 @@ v4l2_stop_streaming(struct v4l2_handle *h)
     for (i = 0; i < h->reqbufs.count; i++) {
 	if (0 != h->buf_me[i].refcount)
 	    ng_waiton_video_buf(&h->buf_me[i]);
+	if (ng_debug)
+	    print_bufinfo(&h->buf_v4l2[i]);
 	if (-1 == munmap(h->buf_me[i].data,h->buf_me[i].size))
 	    perror("munmap");
     }
@@ -1103,7 +873,7 @@ v4l2_stop_streaming(struct v4l2_handle *h)
     /* turn on preview (if needed) */
     if (h->ov_on != h->ov_enabled) {
 	h->ov_on = h->ov_enabled;
-	xioctl(h->fd, VIDIOC_PREVIEW, &h->ov_on, 0);
+	xioctl(h->fd, VIDIOC_OVERLAY, &h->ov_on, 0);
 	if (ng_debug)
 	    fprintf(stderr,"v4l2: overlay on (stop_streaming)\n");
     }
@@ -1118,13 +888,16 @@ v4l2_setformat(void *handle, struct ng_video_fmt *fmt)
 {
     struct v4l2_handle *h = handle;
     
-    h->fmt_v4l2.type = V4L2_BUF_TYPE_CAPTURE;
+    h->fmt_v4l2.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
     h->fmt_v4l2.fmt.pix.pixelformat  = xawtv_pixelformat[fmt->fmtid];
-    h->fmt_v4l2.fmt.pix.flags        = V4L2_FMT_FLAG_INTERLACED;
-    h->fmt_v4l2.fmt.pix.depth        = ng_vfmt_to_depth[fmt->fmtid];
     h->fmt_v4l2.fmt.pix.width        = fmt->width;
     h->fmt_v4l2.fmt.pix.height       = fmt->height;
-    h->fmt_v4l2.fmt.pix.bytesperline = fmt->bytesperline;
+    h->fmt_v4l2.fmt.pix.field        = V4L2_FIELD_ANY;
+    //h->fmt_v4l2.fmt.pix.field        = V4L2_FIELD_ALTERNATE;
+    if (fmt->bytesperline != fmt->width * ng_vfmt_to_depth[fmt->fmtid]/8)
+	h->fmt_v4l2.fmt.pix.bytesperline = fmt->bytesperline;
+    else
+	h->fmt_v4l2.fmt.pix.bytesperline = 0;
 
     if (-1 == xioctl(h->fd, VIDIOC_S_FMT, &h->fmt_v4l2, EINVAL))
 	return -1;
@@ -1149,7 +922,7 @@ v4l2_setformat(void *handle, struct ng_video_fmt *fmt)
 
 /* start/stop video */
 static int
-v4l2_startvideo(void *handle, int fps, int buffers)
+v4l2_startvideo(void *handle, int fps, unsigned int buffers)
 {
     struct v4l2_handle *h = handle;
 
@@ -1159,7 +932,7 @@ v4l2_startvideo(void *handle, int fps, int buffers)
     h->first = 1;
     h->start = 0;
 
-    if (h->cap.flags & V4L2_FLAG_STREAMING)
+    if (h->cap.capabilities & V4L2_CAP_STREAMING)
 	return v4l2_start_streaming(h,buffers);
     return 0;
 }
@@ -1173,7 +946,7 @@ v4l2_stopvideo(void *handle)
 	fprintf(stderr,"v4l2_stopvideo: oops: fps==0\n");
     h->fps = 0;
 
-    if (h->cap.flags & V4L2_FLAG_STREAMING)
+    if (h->cap.capabilities & V4L2_CAP_STREAMING)
 	v4l2_stop_streaming(h);
 }
 
@@ -1185,7 +958,7 @@ v4l2_nextframe(void *handle)
     struct ng_video_buf *buf = NULL;
     int rc,size,frame = 0;
 
-    if (h->cap.flags & V4L2_FLAG_STREAMING) {
+    if (h->cap.capabilities & V4L2_CAP_STREAMING) {
 	v4l2_queue_all(h);
 	frame = v4l2_waiton(h);
 	if (-1 == frame)
@@ -1193,7 +966,7 @@ v4l2_nextframe(void *handle)
 	h->buf_me[frame].refcount++;
 	buf = &h->buf_me[frame];
 	memset(&buf->info,0,sizeof(buf->info));
-	buf->info.ts = h->buf_v4l2[frame].timestamp;
+	buf->info.ts = ng_tofday_to_timestamp(&h->buf_v4l2[frame].timestamp);
     } else {
 	size = h->fmt_me.bytesperline * h->fmt_me.height;
 	buf = ng_malloc_video_buf(&h->fmt_me,size);
@@ -1230,8 +1003,15 @@ v4l2_getimage(void *handle)
 
     size = h->fmt_me.bytesperline * h->fmt_me.height;
     buf = ng_malloc_video_buf(&h->fmt_me,size);
-    if (h->cap.flags & V4L2_FLAG_READ) {
+    if (h->cap.capabilities & V4L2_CAP_READWRITE) {
 	rc = read(h->fd,buf->data,size);
+	if (-1 == rc  &&  EBUSY == errno  &&  h->ov_on) {
+	    h->ov_on = 0;
+	    xioctl(h->fd, VIDIOC_OVERLAY, &h->ov_on, 0);
+	    rc = read(h->fd,buf->data,size);
+	    h->ov_on = 1;
+	    xioctl(h->fd, VIDIOC_OVERLAY, &h->ov_on, 0);
+	}
 	if (rc != size) {
 	    if (-1 == rc) {
 		perror("v4l2: read");
@@ -1262,5 +1042,5 @@ v4l2_getimage(void *handle)
 extern void ng_plugin_init(void);
 void ng_plugin_init(void)
 {
-    ng_vid_driver_register(NG_PLUGIN_MAGIC,PLUGNAME,&v4l2_driver);
+    ng_vid_driver_register(NG_PLUGIN_MAGIC,__FILE__,&v4l2_driver);
 }
